# Module for Fourier interpolation of 2D functions sampled on a polar grid
# Author: A. Corstanje (a.corstanje@astro.ru.nl), 2020 - 2023

import numpy as np
import scipy.interpolate as intp

class interp2d_fourier:
    @classmethod
    def get_ordering_indices(self, x, y):
        """ Produces ordering indices to create (radius, phi) 2D-array from unordered x and y (1D-)arrays.

        Parameters
        ----------
        x : 1D array of x positions
        y : 1D array of y positions
        """

        radius = np.sqrt(x**2 + y**2)
        phi = np.arctan2(y, x) # uses interval -pi..pi
        phi = np.deg2rad(np.around(np.rad2deg(phi)))
        phi[np.abs(phi) < np.deg2rad(0.01)] = 0.0 # avoid pathology
        phi[phi<0] += 2*np.pi # put into 0..2pi for ordering
        phi_sorting = np.argsort(phi)

        # Assume star-shaped pattern, i.e. radial # steps = number of (almost) identical phi-values
        # May not work very near (0, 0)
        self._phi0 = phi[phi_sorting][0]
        #print('Phi0 = %1.8f' % self._phi0)
        test = phi[phi_sorting] - self._phi0
        radial_steps = len(np.where(np.abs(test) < 0.0001)[0])
        phi_steps = len(phi_sorting) // radial_steps
        phi_sorting = phi_sorting.reshape((phi_steps, radial_steps))

        indices = np.argsort(radius[phi_sorting], axis=1)
        for i in range(phi_steps): # Sort by radius; should be possible without for-loop...
            phi_sorting[i] = phi_sorting[i][indices[i]]
        ordering_indices = phi_sorting.T # get shape (radial_steps, phi_steps)

        return ordering_indices

    @classmethod
    def cos_sin_components(cls, fourier):
        """
        Convert complex FFT as from np.fft.rfft to real-valued cos, sin components
        Input: complex Fourier components, with Fourier series running along the last axis.
        """
        cos_components = 2*np.real(fourier)
        cos_components[..., 0] *= 0.5
        cos_components[..., -1] *= 0.5
        sin_components = -2*np.imag(fourier)

        return (cos_components, sin_components)

    def __init__(self, x, y, values, radial_method='cubic', fill_value=None, recover_concentric_rings=False):
        """
        # Produce a callable instance (given by the function __call__) to interpolate a function value(x, y) sampled at the input positions (x, y)

        Parameters
        ----------
        x : 1D array of x positions of simulated antennas
        y : idem for y
        values : the function values (as 1D array) at positions (x, y)
        radial_method : the interp1d method for interpolating Fourier components along the radial axis (optional, default 'cubic' i.e. cubic splines)
        fill_value : the fill value to pass to interp1d to use for a radius outside the min..max radius interval from the input. Set to 'extrapolate' to extrapolate beyond radial limits; accuracy outside the interval is limited. Optional, default None, meaning stay constant for r < r_min, and 0 for r > r_max
        recover_concentric_rings : set True if the grid is not purely circular-symmetric; results may not be accurate. Optional, default False
        """
        # Convert (x, y) to (r, phi), make 2d position array, sorting positions and values by r and phi
        radius = np.sqrt(x**2 + y**2)

        ordering_indices = self.get_ordering_indices(x, y)
        values_ordered = np.copy(values)[ordering_indices]

        # Store the (unique) radius values
        self.radial_axis = radius[ordering_indices][:, 0]
        # Check if the radius does not vary along angular direction (with tolerance)
        if np.max(np.std(radius[ordering_indices], axis=1)) > 0.1 * np.min(radius):
            if not recover_concentric_rings:
                raise ValueError("Radius must be (approx.) constant along angular direction." + \
                                " You can try to \"fix\" that by using \"recover_concentric_rings=True\"")
            else:
                self.radial_axis = np.mean(radius[ordering_indices], axis=1)
                values_ordered_interpolated = []
                for x, y in zip(radius[ordering_indices].T, values_ordered.T):
                    intpf = intp.interp1d(
                        x, y, axis=0, kind=radial_method, fill_value='extrapolate')
                    values_ordered_interpolated.append(intpf(self.radial_axis))
                values_ordered = np.array(values_ordered_interpolated).T

        # FFT over the angular direction, for each radius
        self.angular_FFT = np.fft.rfft(values_ordered, axis=1)
        length = values_ordered.shape[-1]
        self.angular_FFT /= float(length) # normalize

        # Produce interpolator function, interpolating the FFT components as a function of radius
        self.interpolator_radius = intp.interp1d(self.radial_axis, self.angular_FFT, axis=0, kind=radial_method, fill_value='extrapolate') # Interpolates the Fourier components along the radial axis


    def __call__(self, x, y, max_fourier_mode=None):
        """
        Interpolate the input used in __init__ for input positions (x, y)

        Parameters
        ----------
        x : x positions as float or numpy ND array
        y : idem for y
        max_fourier_mode : optional cutoff for spatial frequencies along circles, i.e. do Fourier sum up to (incl.) this mode. Default None i.e. do all modes
        """
        radius = np.sqrt(x**2 + y**2)
        phi = np.arctan2(y, x) - self._phi0

        # Interpolate Fourier components over all values of radius
        fourier = self.interpolator_radius(radius)
        fourier_len = fourier.shape[-1]

        (cos_components, sin_components) = interp2d_fourier.cos_sin_components(fourier)

        limit = max_fourier_mode+1 if max_fourier_mode is not None else fourier_len
        mult = np.linspace(0, limit-1, limit).astype(int) # multipliers for Fourier modes, as k in cos(k*phi), sin(k*phi)
        result = np.zeros_like(radius)
        if isinstance(phi, float):
            result = np.sum(cos_components[..., 0:limit] * np.cos(phi * mult)) + np.sum(sin_components[..., 0:limit] * np.sin(phi * mult))
        else:
            result = np.sum(cos_components[..., 0:limit] * np.cos(phi[..., np.newaxis] * mult), axis=-1) + np.sum(sin_components[..., 0:limit] * np.sin(phi[..., np.newaxis] * mult), axis=-1)
        # The Fourier sum done explicitly, as sum_k( c_k cos(k phi) + s_k sin(k phi) )

        return result

    # Some getters for the angular FFT, its radial interpolator function, and the radial axis points used

    def get_angular_FFT(self):

        return self.angular_FFT

    def get_angular_FFT_interpolator(self):

        return self.interpolator_radius

    def get_radial_axis(self):

        return self.radial_axis



# Separate module for cross-correlating and making a demo plot
# Author: A. Corstanje, (a.corstanje@astro.ru.nl), 2023

import h5py
import numpy as np
import matplotlib.pyplot as plt
from scipy import signal as scipy_signal

def get_freq_axis(signal):
    """
    Return the frequency axis of a (real) FFT spectrum of time series 'signal' as 1D array
    Frequencies in MHz (hence the 1.0e-6), assumed a 0.1 ns sample period
    """
    return (1.0e-6 * np.fft.rfftfreq(signal.shape[0], d=0.1e-9)) # MHz

def do_filter_signal_lowpass(signal, cutoff_freq):
    """
    For one signal time series, do lowpass filtering.

    Parameters
    ----------
    signal : time series, 1D array
    cutoff_freq : high frequency cutoff in MHz
    """
    freqs = get_freq_axis(signal)

    filter_indices = np.where(freqs > cutoff_freq)
    spectrum = np.fft.rfft(signal)
    spectrum[filter_indices] *= 0.0
    signal_filtered = np.fft.irfft(spectrum)

    return signal_filtered

def get_crosscorrelation(test_signal, orig_signal, upsampling_factor=10):
    """
    Get normalized cross-correlation between 'test_signal' and 'orig_signal', returned as 'CC_zeroshift'.
    Also returns: normalized cross-correlation optimized over time shift between the two signals;
    time difference for which the cross-correlation is maximal;
    relative energy difference

    Parameters
    ----------
    test_signal : time series, 1D array
    orig_signal : idem
    upsampling_factor : upsampling factor for sub-sample timing accuracy, used to obtain optimal cross-correlation (optimized over arrival times). Default 10.
    """

    orig_signal_upsampled = scipy_signal.resample(orig_signal, upsampling_factor*len(orig_signal) )
    test_signal_upsampled = scipy_signal.resample(test_signal, upsampling_factor*len(test_signal) )

    #orig_upsampled = scipy.signal.resample(orig_signal, upsampling_factor*len(orig_signal) )
    #interp_upsampled = scipy.signal.resample(interpolated_signal, upsampling_factor*len(interpolated_signal) )

    crosscorr = scipy_signal.correlate(test_signal_upsampled, orig_signal_upsampled)
    #lags = signal.correlation_lags(orig_pulse.size, interpolated_pulse.size)
    #    lag = lags[np.argmax(crosscorr)]
    normalization = np.sqrt(np.sum(orig_signal_upsampled**2) * np.sum(test_signal_upsampled**2))
    crosscorr /= normalization

    autocorr = scipy_signal.correlate(orig_signal_upsampled, orig_signal_upsampled)
    max_autocorr = np.argmax(autocorr) # this is at "t=0"

    # Get Delta t, maximum at zero time shift, maximum overall
    CC_optimized_timeshift = np.max(crosscorr)
    CC_zeroshift = crosscorr[max_autocorr] # the CC value at fixed timing

    delta_t = 0.1 * (1.0 / upsampling_factor) * (np.argmax(crosscorr) - np.argmax(autocorr)) # 0.1 ns per sample in original signal

    # so delta_t is given in ns

    orig_energy = np.sum(orig_signal_upsampled**2)
    test_energy = np.sum(test_signal_upsampled**2)
    energy_rel_diff = (test_energy - orig_energy) / orig_energy

    return (CC_zeroshift, CC_optimized_timeshift, delta_t, energy_rel_diff)



def plot_pulse_and_spectrum(orig_pulse, interpolated_pulse, x, y, cutoff_freq, pol):
    """
    Plots an interpolated pulse together with a 'true' simulated pulse

    Parameters
    ----------
    orig_pulse : time series, 1D array
    interpolated_pulse : idem
    x : the x position (float), for annotation in the plot
    y : idem for y
    cutoff_freq : value of estimated cutoff frequency, for annotation only
    pol : polarization number
    """
    time_axis = 0.1 * np.arange(len(orig_pulse))
    radius = np.sqrt(x**2 + y**2)
    freqs = get_freq_axis(orig_pulse)

    (CC_zeroshift, CC_optimized_timeshift, delta_t, energy_rel_diff) = get_crosscorrelation(orig_pulse, interpolated_pulse)

    fig, ax = plt.subplots(figsize=(10.67, 4), nrows=1, ncols=2)
    ax1, ax2 = ax[0], ax[1]

    #plt.figure()
    ax1.plot(time_axis, 1.0e6 * orig_pulse, label='orig pulse', lw=2)
    ax1.plot(time_axis, 1.0e6 * interpolated_pulse, label='interpolated pulse', lw=2)
    residual = 1.0e6 * (interpolated_pulse - orig_pulse)
    ax1.plot(time_axis, residual, label='difference', lw=2, c='g')

    interpolated_energy = np.sum(interpolated_pulse**2)
    orig_energy = np.sum(orig_pulse**2)
    #fixed_pulse = interpolated_pulse * np.sqrt(orig_energy / interpolated_energy)
    #plt.plot(time_axis, fixed_pulse, label='fixed pulse', c='r', lw=1)
    ax1.grid()
    ax1.set_xlabel('Time [ ns ]')
    ax1.set_ylabel(r'E-field [ $\mu$V/m ]')
    ax1.set_xlim(0.0, 50.0)
    ax1.legend(loc='best')

    #plt.figure()
    orig_pulse_powerspec = np.abs(np.fft.rfft(orig_pulse))**2
    interp_pulse_powerspec = np.abs(np.fft.rfft(interpolated_pulse))**2
    ax2.plot(freqs, orig_pulse_powerspec, label='Orig pulse')
    ax2.plot(freqs, interp_pulse_powerspec, label='Interpolated pulse')
    ax2.text(0.98, 0.40, 'Position x = %3.1f, y = %3.1f, r = %3.2f m, pol = %d' % (x, y, radius, pol), transform=plt.gca().transAxes, ha='right') #, va='right')
    ax2.text(0.98, 0.30, 'CC = %1.5f, CC_max = %1.5f' % (CC_zeroshift, CC_optimized_timeshift), transform=plt.gca().transAxes, ha='right')
    ax2.text(0.98, 0.20, 'delta_t = %1.2f ns, cutoff freq = %3.1f MHz' % (delta_t, cutoff_freq), transform=plt.gca().transAxes, ha='right')

    #plt.yscale('log')
    ax2.grid()
    ax2.set_xlabel('Frequency [ MHz ]')
    ax2.set_ylabel('Power spectrum [ a.u. ]')
    ax2.set_xlim(0, 500)
    ax2.set_ylim(0.0, 1.2*np.max(orig_pulse_powerspec))
    ax2.legend(loc='best')


def read_data_hdf5(filename):
    """
    Reading in the demo data hdf5 file.
    The time traces inside are CoREAS E-fields, converted to 2 'on-sky' polarizations.
    """
    try:
        demo_file = h5py.File(filename, 'r')
    except:
        raise ValueError('Cannot read data file; demo data downloaded with download_demo_data.sh?')
    zenith = demo_file.get('zenith')[()]
    azimuth = demo_file.get('azimuth')[()]
    xmax = demo_file.get('xmax')[()]
    footprint_positions = np.array(demo_file.get('footprint_positions'))
    test_positions = np.array(demo_file.get('test_positions'))
    (footprint_pos_x, footprint_pos_y) = (footprint_positions[:, 0], footprint_positions[:, 1])
    (test_pos_x, test_pos_y) = (test_positions[:, 0], test_positions[:, 1])

    footprint_antenna_data = np.array(demo_file.get('footprint_antennas'))
    test_antenna_data = np.array(demo_file.get('test_antennas'))
    demo_file.close()

    return (zenith, azimuth, xmax, footprint_pos_x, footprint_pos_y, test_pos_x, test_pos_y, footprint_antenna_data, test_antenna_data)



# Minimal usage demo script of Fourier interpolation of pulse signals along simulated CR radio footprints
# Author: A. Corstanje (a.corstanje@astro.ru.nl), 2023

import numpy as np
import matplotlib.pyplot as plt
plt.ion()

import signal_interpolation_fourier as sigF

import demo_helper

""" For using the interpolator, the x and y positions of the simulated antennas in the shower plane are needed,
along with the time traces for all these antennas and polarizations
The antennas need not be ordered by position

Shapes are 1D array for x, y (in meters)
antenna traces 'footprint_antenna_data' in shape (Nant, Nsamples, Npol), i.e., in this example (208, 4082, 2)
Additional antenna positions have been simulated ('test_antenna_data' and 'test_pos_x', 'test_pos_y') to test the interpolation accuracy

When within < ~10 degrees from the North-South axis (including zenith), rotate the `on-sky` polarizations by 45 degrees
to avoid alignment with vx(vxB) and therefore having near-zero signals along circles in the footprint
"""
# Read in demo data
demo_filename = 'demo_shower.h5'
(zenith, azimuth, xmax, footprint_pos_x, footprint_pos_y, test_pos_x, test_pos_y, footprint_antenna_data, test_antenna_data) = demo_helper.read_data_hdf5(demo_filename)

"""
Initialize the interpolator object. It needs the antenna positions (x and y as 1D arrays) and their time traces as 3D arrays (Nants, Nsamples, Npols)
By default the phase-interpolating method is "phasor" (see article)
"""
print('Initializing interpolator object...')
phase_method = "phasor" # other option is "timing" which takes longer to initialize (to check which performs better in use cases not yet tested)
signal_interpolator = sigF.interp2d_signal(footprint_pos_x, footprint_pos_y, footprint_antenna_data, verbose=False, phase_method=phase_method)

test_index = 124 # take one of the 250 test positions
pol = 0
this_x, this_y = test_pos_x[test_index], test_pos_y[test_index]
print('Interpolating pulse at position x = %3.2f, y = %3.2f m' % (this_x, this_y))

orig_pulse = test_antenna_data[test_index] # shape is (Nsamples, Npol)

"""
Call the interpolator object to obtain the interpolated pulse at the desired position
Return shape is (Nsamples, Npol)
Optionally, it can be low-pass filtered to an estimated reliable cutoff frequency
This is a reliable yet sometimes overly conservative estimate up to which frequency the interpolation is accurate
"""
interpolated_pulse = signal_interpolator(this_x, this_y) #, filter_up_to_cutoff=True)

# Make a plot of the trace and spectrum
demo_helper.plot_pulse_and_spectrum(orig_pulse[:, pol], interpolated_pulse[:, pol], this_x, this_y, -1, pol)



# Demo script of Fourier interpolation of pulse signals along simulated CR radio footprints
# Author: A. Corstanje (a.corstanje@astro.ru.nl), 2023

import numpy as np
import matplotlib.pyplot as plt
plt.ion()

import signal_interpolation_fourier as sigF

import demo_helper

"""
For using the interpolator, the x and y positions of the simulated antennas in the shower plane are needed,
along with the time traces for all these antennas and polarizations
Shapes are 1D array for x, y (in meters)
antenna traces in shape (Nant, Nsamples, Npol), i.e., in this example (208+250, 4082, 2)
"""

demo_filename = 'demo_shower.h5'
(zenith, azimuth, xmax, footprint_pos_x, footprint_pos_y, test_pos_x, test_pos_y, footprint_antenna_data, test_antenna_data) = demo_helper.read_data_hdf5(demo_filename)

nof_test_positions = test_pos_x.shape[0] # the number of test antennas, here 250
azimuth_deg = (azimuth % (2*np.pi)) * 180.0/np.pi
azimuth_deg_clockwise_from_north = 90.0 - azimuth_deg
zenith_deg = zenith * 180.0/np.pi

print('Shower data from a 10^17 proton, azimuth = %3.1f, zenith = %3.1f deg, Xmax = %4.2f g/cm2' % (azimuth_deg_clockwise_from_north, zenith_deg, xmax))
print('(azimuth as clockwise from North)')

test_radius = np.sqrt(test_pos_x**2 + test_pos_y**2) # the core distance of each of the test antennas

phase_method = "phasor" # other option is "timing"
if phase_method == "timing":
    print('Initializing interpolator, this may take 1 to 3 minutes...')
else:
    print('Initializing interpolator...')

signal_interpolator = sigF.interp2d_signal(footprint_pos_x, footprint_pos_y, footprint_antenna_data, verbose=True, phase_method=phase_method)
print('Done.')

test_indices = (23, 124, 20, 34) # Evaluate interpolation at these test positions
pol = 0 # polarization '0' for demo plots
for index in test_indices:
    this_x, this_y = test_pos_x[index], test_pos_y[index]
    print('Interpolating pulse at position x = %3.2f, y = %3.2f m' % (this_x, this_y))

    orig_pulse = test_antenna_data[index]

    interpolated_pulse = signal_interpolator(this_x, this_y) #, filter_up_to_cutoff=True)

    orig_pulse = orig_pulse[:, pol]
    interpolated_pulse = interpolated_pulse[:, pol] # do only strongest polarization

    this_cutoff_freq = signal_interpolator.get_cutoff_freq(this_x, this_y, pol)

    #filtered_orig = demo_helper.do_filter_signal_lowpass(orig_pulse, this_cutoff_freq)

    (CC_zeroshift, CC_optimized_timeshift, delta_t, energy_rel_diff) = demo_helper.get_crosscorrelation(orig_pulse, interpolated_pulse)
    print('Normalized cross correlation (CC) = %1.4f, time mismatch = %1.3f ns' % (CC_zeroshift, delta_t))

    demo_helper.plot_pulse_and_spectrum(orig_pulse, interpolated_pulse, this_x, this_y, this_cutoff_freq, pol)

"""
Evaluate cross-correlation between true and interpolated pulses
for all 250 test positions, and for each of the 2 `on-sky' polarizations
"""
CC_values = np.zeros( (nof_test_positions, 2) )
distances = np.zeros(nof_test_positions)
for index in range(nof_test_positions):
    this_x, this_y = test_pos_x[index], test_pos_y[index]
    core_distance = np.sqrt(this_x**2 + this_y**2)
    print('Interpolating pulse at position x = %3.2f, y = %3.2f m' % (this_x, this_y))

    orig_pulse = test_antenna_data[index]

    interpolated_pulse = signal_interpolator(this_x, this_y)

    for pol in (0, 1):
        this_cutoff_freq = signal_interpolator.get_cutoff_freq(this_x, this_y, pol)

        #filtered_orig = demo_helper.do_filter_signal_lowpass(orig_pulse, this_cutoff_freq)

        (CC_zeroshift, CC_optimized_timeshift, delta_t, energy_rel_diff) = demo_helper.get_crosscorrelation(orig_pulse[:, pol], interpolated_pulse[:, pol])
        print('Normalized cross correlation (CC) = %1.4f, time mismatch = %1.3f ns' % (CC_zeroshift, delta_t))

        CC_values[index, pol] = CC_zeroshift
        distances[index] = core_distance

plt.figure()
plt.scatter(distances, CC_values[:, 0], label='pol 0')
plt.scatter(distances, CC_values[:, 1], label='pol 1')
plt.xlabel('Core distance [ m ]')
plt.ylabel('Normalized CC')
plt.grid()
plt.legend(loc='best')



# Module for Fourier interpolation of pulsed signals along simulated radio footprints of cosmic-ray air showers
# Author: A. Corstanje (a.corstanje@astro.ru.nl), 2023
#
# See article: arXiv:xxxx.xxxxx, doi xxx
# Please cite this when using code and/or methods in your analysis

import numpy as np
from scipy.signal import hilbert, resample
import interpolation_fourier as interpF

class interp2d_signal:

    def get_spectra(self, signals):
        """
        # Do FFT of 'signals', assumed shape (Nants, Nsamples, Npol) i.e. the time traces are along the second axis.
        # Produce absolute-amplitude spectrum and phase spectrum
        Parameters
        ----------
        signals : the input time traces
        """
        Nsamples = signals.shape[1]

        if self.verbose:
            print('Doing FFTs...', end=' ')
        all_antennas_spectrum = np.fft.rfft(signals, axis=1)
        abs_spectrum = np.abs(all_antennas_spectrum)
        phasespectrum = np.angle(all_antennas_spectrum)
        unwrapped_phases = np.unwrap(phasespectrum, axis=1, discont=0.7*np.pi)
        if self.verbose:
            print('done.')

        #antnr = 1555
        #print(pos_sim_UVW[antnr])
        freqs = np.fft.rfftfreq(Nsamples, d=self.sampling_period)
        freqs /= 1.0e6 # in MHz

        freqstep = freqs[1] - freqs[0]
        nof_freq_channels = len(np.where( (freqs < 500) )[0])
        if self.verbose:
            print('Frequency channel width %2.4f MHz, there are %d channels between 0 and 500 MHz' % (freqstep, nof_freq_channels))

        return (freqs, all_antennas_spectrum, abs_spectrum, phasespectrum, unwrapped_phases)

    def hilbert_envelope_timing(self, signals, lowfreq=30.0, highfreq=500.0, upsample_factor=10, sum_over_pol=True, do_hilbert_envelope=True):
        """
        Produce pulse arrival times from Hilbert envelope maxima (if do_hilbert_envelope) or from raw E-field maxima.
        Assumed shape for 'signals' is (Nant, Nsamples, Npols), i.e. time traces along second axis
        Filtering is done between lowfreq and highfreq in MHz, timing is done on filtered signals (after inverse-FFT)
        Option sum_over_pol is to sum the square of the Hilbert envelopes of each polarization, to get one arrival time over all polarizations (may help at low signal strength), default True.

        Parameters
        ----------
        signals : 3D array of shape (Nant, Nsamples, Npols)
        lowfreq : low-frequency cutoff, default 30.0 MHz
        highfreq : high-frequency cutoff, default 500.0 MHz
        upsample_factor : upsampling factor to use for sub-sample timing accuracy, default 10
        sum_over_pol : sum over polarizations if do_hilbert_envelope, default True
        do_hilbert_envelope : use Hilbert envelope for timing if True, else direct E-field maximum
        """
        (Nant, Nsamples, Npols) = signals.shape

        if self.verbose:
            print('Bandpass filtering %d to %d MHz' % (int(lowfreq), int(highfreq)))
        spectrum = np.fft.rfft(signals, axis=1)
        freqs = np.fft.rfftfreq(Nsamples, d=self.sampling_period)
        freqs /= 1.0e6 # in MHz
        filtering_out = np.where( (freqs < lowfreq) | (freqs > highfreq)) # or!

        spectrum[:, filtering_out, :] *= 0.0
        filtered_signals = np.fft.irfft(spectrum, axis=1)

        # Get strongest polarization
        power_per_pol = np.sum(np.sum(filtered_signals**2, axis=1), axis=0)
        strongest_pol = np.argmax(power_per_pol)
        if self.verbose:
            print('Strongest polarization is %d' % strongest_pol)

        timestep = self.sampling_period # in s
        if self.verbose:
            print('Upsampling by a factor %d' % upsample_factor)
        signals_upsampled = resample(filtered_signals, upsample_factor*Nsamples, axis=1)

        nof_samples = signals_upsampled.shape[1]
        signals_upsampled = np.roll(signals_upsampled, nof_samples//2, axis=1) # Put in the middle of the block, avoiding negative values in timing

        if do_hilbert_envelope:
            if self.verbose:
                print('Hilbert envelope')
            hilbert_envelope = np.abs(hilbert(signals_upsampled, axis=1))

            if sum_over_pol:
                hilbert_sum_over_pol = np.sqrt( np.sum(hilbert_envelope**2, axis=2) )

                pulse_timings = (np.argmax(hilbert_sum_over_pol, axis=1) - nof_samples//2) * (timestep/upsample_factor)
            else:
                pulse_timings_per_pol = (np.argmax(hilbert_envelope, axis=1) - nof_samples//2) * (timestep/upsample_factor)

        else:
            pulse_timings_per_pol = (np.argmax(signals_upsampled, axis=1) - nof_samples//2) * (timestep/upsample_factor)

        #pulse_timings = np.argmax(hilbert_envelope[:, :, 1], axis=1) * (timestep/upsample_factor)
        if self.verbose:
            print('Timings done')

        if do_hilbert_envelope and sum_over_pol:
            pulse_timings_per_pol = np.zeros( (Nant, Npols) )
            for pol in range(Npols):
                pulse_timings_per_pol[:, pol] += pulse_timings # want to do this without yucky for loop

        return pulse_timings_per_pol


    def phase_wrap(self, phases):
        """
        wrap 'phases' (float or any array shape) into interval (-pi, pi)
        """
        return ((phases + np.pi) % (2 * np.pi) - np.pi)

    def timing_corrected_phases(self, freqs, phase_spectrum, pulse_timings):
        """
        Take phase_spectrum as input
        Account for a linear function from the pulse_timings, i.e.
        according to delta_phi = 2 pi f delta_t
        Return as timing-corrected phase spectrum

        Parameters
        ----------
        freqs : frequency axis of FFTs
        phase_spectrum : 3D array of shape (Nants, Nphases, Npols) containing phase spectra
        pulse_timings : 2D array of shape (Nants, Npols) containing pulse timings
        """
        (Nants, Nphases, Npols) = phase_spectrum.shape

        freq_step = freqs[1]
        phase_spectrum_corrected = np.zeros( phase_spectrum.shape )

        for i, ant in enumerate(range(Nants)):
            for pol in range(Npols):

                this_phase_corrections = 2*np.pi * (freqs*1.0e6) * pulse_timings[i, pol] # plus or minus sign?
                phase_spectrum_corrected[ant, :, pol] = phase_spectrum[ant, :, pol] + this_phase_corrections # this can be done more efficiently...

        # WRAP phases into 0..2*pi
        phase_spectrum_corrected = self.phase_wrap(phase_spectrum_corrected)
        phase_spectrum_corrected = np.unwrap(phase_spectrum_corrected, axis=1, discont=0.7*np.pi)
        # needed?

        return phase_spectrum_corrected

    def phase_unwrap_2d(self, x, y, phases):
        """
        Basic method to unwrap phases in 2D
        The general problem of optimal 2D phase unwrapping is NP-complete (see ref in article), so this will only work for well-behaved, slowly varying phases
        First do 1D unwrap over radial directions
        Then 1D unwrap over angular directions (along circles in the radial grid)

        Parameters
        ----------
        x : 1D array of antenna position x (in any order)
        y : same for y (same order)
        phases : 1D array, i.e. one phase per antenna (same order of antennas as x, y)
        """
        # Put them into a 2D array with a radial and an angular axis
        indices = interpF.interp2d_fourier.get_ordering_indices(x, y)
        phases_ordered = phases[indices]

        # First along radial axis, then angular axis
        phases_unwrapped_2d = np.unwrap(phases_ordered, axis=0)
        phases_unwrapped_2d  = np.unwrap(phases_unwrapped_2d, axis=1)
        #phases_unwrapped_2d  = np.unwrap(phases_unwrapped_2d, axis=0)

        # Back into original 1D shape...
        phases_unwrapped = np.zeros( phases.shape )

        (Nradial, Nangular, Npol) = phases_unwrapped_2d.shape

        phases_unwrapped[indices.ravel()] = phases_unwrapped_2d.reshape( (Nradial*Nangular, Npol) )

        return phases_unwrapped

    def degree_of_coherency(self, low_freq=30.0, high_freq=500.0):
        """
        This implements Eq. (2.4) in the article, for given frequency band limits
        """
        complex_phases = np.exp(1.0j * self.phasespectrum_corrected)
        ampli = self.abs_spectrum

        spectrum_corrected = ampli * complex_phases

        freq_range = np.where( (self.freqs > low_freq) & (self.freqs < high_freq))[0]

        complex_sum = np.sum(spectrum_corrected[:, freq_range, :], axis=1)

        abs_sum = np.sum(ampli[:, freq_range, :], axis=1)

        coherency = np.abs(complex_sum) / abs_sum

        return coherency

    def get_constant_phases(self, low_freq=30.0, high_freq=500.0):
        """
        This implements Eq. (2.3) in the article, for given frequency band limits.
        Phases have been corrected to have maximum Hilbert envelope at "t"=0
        So, add up the complex phases, weighted by the amplitudes, to get the constant phase
        which determines if the pulse is cos-like or sin-like, or a value in between
        """
        complex_phases = np.exp(1.0j * self.phasespectrum_corrected)
        ampli = self.abs_spectrum

        spectrum_corrected = ampli * complex_phases

        freq_range = np.where( (self.freqs > low_freq) & (self.freqs < high_freq))[0]
        complex_sum = np.sum(spectrum_corrected[:, freq_range, :], axis=1)

        const_phases = np.angle(complex_sum)

        # Do unwrapping in 2D to avoid 2 pi periodicity mismatches (spurious jumps of 2 pi)
        const_phases_unwrapped = self.phase_unwrap_2d(self.pos_x, self.pos_y, const_phases)

        return const_phases_unwrapped

    def get_coherency_vs_frequency(self, bandwidth=50.0, low_freq=30.0, high_freq=500.0, coherency_cutoff=0.8):
        """
        Compute `degree of coherency` as from Eq. (2.4) in a sliding frequency window
        each with bandwidth 50 MHz (or optional value given)
        Then, see at which frequency this value first drops below 'coherency_cutoff'
        That frequency defines the 'cutoff frequency' (cutoff_freq) below which
        the signal is well approximated by a standard impulse, and thus reliably interpolated.

        Parameters
        ----------
        bandwidth : frequency bandwidth in MHz, default 50.0
        low_freq : low frequency cutoff in MHz, default 30.0
        high_freq : high frequency cutoff in MHz, default 500.0
        coherency_cutoff : threshold value for coherency level, default 0.8
        """
        freq_indices = np.where( (self.freqs > low_freq) & (self.freqs < (high_freq)))[0] # high_freq - bandwidth?

        (Nants, Nsamples, Npols) = self.phasespectrum_corrected.shape
        coherency_vs_freq = np.zeros( (Nants, len(freq_indices), Npols))

        cutoff_freq = np.zeros( (Nants, Npols) )

        for i, freq_index in enumerate(freq_indices):
            if self.verbose:
                print('%d / %d' % (i, len(freq_indices)))
            this_freq = self.freqs[freq_index]
            coherency = self.degree_of_coherency(low_freq=this_freq, high_freq=this_freq+bandwidth)
            coherency_vs_freq[:, i, :] = coherency

        # cutoff_values = np.where(coherency_vs_freq < 0.8)
        # TODO do this without / with fewer for loops
        # do explicitly in for loops for now...
        for ant in range(Nants):
            for pol in range(Npols):
                cutoff_index = np.where(coherency_vs_freq[ant, :, pol] < coherency_cutoff)[0] # first index where < 0.8
                if len(cutoff_index) > 0:
                    cutoff_index = cutoff_index[0]
                else:
                    cutoff_index = -1
                cutoff_freq[ant, pol] = min(self.freqs[freq_indices[cutoff_index]], high_freq) # max 500.0 cap, needed?

        return (coherency_vs_freq, cutoff_freq)

    def get_freq_dependent_timing_correction(self, lowfreq=30.0, highfreq=500.0, bandwidth=50.0, upsample_factor=10, ignore_cutoff_freq_in_timing=False):
        """
        Implements "method (2)" to account for phase spectra towards higher frequencies.
        Determines arrival time (E-field maximum) in a 50 MHz (or 'bandwidth') sliding frequency window, by filtering to each frequency window and inverse-FFT.
        Arrival time is mapped to phase in the center of each window.

        Parameters
        ----------
        low_freq : low frequency cutoff in MHz, default 30.0
        high_freq : high frequency cutoff in MHz, default 500.0
        bandwidth : frequency bandwidth in MHz, default 50.0
        upsample_factor : upsampling factor for timing accuracy, default 10
        ignore_cutoff_freq_in_timing: proceed with pulse timing beyond cutoff frequency, default False
        """
        start_freq = self.freqs[self.freqs>=lowfreq][0]
        end_freq = self.freqs[self.freqs<=(highfreq+bandwidth/2)][-1]

        freq_indices = np.where((self.freqs >= lowfreq) & (self.freqs < (highfreq+bandwidth/2)))[0]

        spectrum_after_correction_so_far = self.abs_spectrum * np.exp(1j * self.phasespectrum_corrected)
        signals = np.fft.irfft(spectrum_after_correction_so_far, axis=1)

        freq_dependent_timing = np.zeros( self.abs_spectrum.shape )

        for i, freq_index in enumerate(freq_indices):
            if i % 3 != 0:
                freq_dependent_timing[:, freq_index, :] += freq_dependent_timing[:, freq_index-1, :]
            else:
                this_freq = self.freqs[freq_index]

                band_low = max(lowfreq, this_freq - bandwidth/2)
                band_high = min(highfreq+bandwidth, this_freq + bandwidth/2)
                #band_low = np.max(lowfreq, this_freq - bandwidth/2)
                #band_high = min(highfreq+bandwidth/2, this_freq + bandwidth/2)
                #band_low = max(lowfreq, band_high - bandwidth)
                if ((band_high - band_low) < bandwidth) and (band_low + bandwidth < band_high):
                    band_high = band_low + bandwidth
                # check bandwidth
                if self.verbose:
                    print('Bandwidth %3.1f MHz, from %3.1f to %3.1f MHz' % (band_high-band_low, band_low, band_high))

                this_timing = self.hilbert_envelope_timing(signals, lowfreq=band_low, highfreq=band_high, upsample_factor=upsample_factor, do_hilbert_envelope=False)
                # returns timings for (Nant, Npol)
                freq_dependent_timing[:, freq_index, :] += this_timing # check index broadcasting

        # do cutoff fixing here
        # keep timing constant beyond frequency cutoff
        if not ignore_cutoff_freq_in_timing:
            (Nants, Nsamples, Npols) = freq_dependent_timing.shape
            for ant in range(Nants):
                for pol in range(Npols):

                    # Keep timing values constant for frequencies > local cutoff value
                    this_index = np.where(self.freqs >= self.cutoff_freq[ant, pol])[0][0]
                    freq_dependent_timing[ant, this_index:, pol] = freq_dependent_timing[ant, this_index-1, pol]

                    for i, index in enumerate(freq_indices):
                        if i == 0:
                            continue
                        # Just in case, a similar stopping criterion when successive timings get too `noisy`
                        if np.abs(freq_dependent_timing[ant, index, pol] - freq_dependent_timing[ant, index-1, pol]) > 0.5e-9:
                            freq_dependent_timing[ant, index:, pol] = freq_dependent_timing[ant, index-1, pol]
                            break

        return freq_dependent_timing

    def nearest_antenna_index(self, x, y, same_radius=False, tolerance=1.0):
        """
        Search for closest antenna at the same radius, within tolerance (m), when same_radius=True
        Otherwise find the nearest antenna to (x, y) and return its index
        The nearest antenna position is then referenced as self.pos_x[index], self.pos_y[index]

        Parameters
        ----------
        x : x position (float, single value)
        y : idem for y
        same_radius : consider only antennas at the same radius, +/- 'tolerance'
        tolerance : tolerance in m for same_radius search
        """
        if not same_radius:
            index = np.argmin( (self.pos_x - x)**2 + (self.pos_y - y)**2 )
        else: # a bit convoluted...
            radius_antennas = np.sqrt(self.pos_x**2 + self.pos_y**2)
            thisradius = np.sqrt(x**2 + y**2)

            indices_same_radius = np.where( (radius_antennas > (thisradius-tolerance)) & (radius_antennas < (thisradius+tolerance)) )[0]
            assert len(indices_same_radius) > 0
            ant_x = self.pos_x[indices_same_radius]
            ant_y = self.pos_y[indices_same_radius]

            index_min = np.argmin( (ant_x - x)**2 + (ant_y - y)**2 )

            index = indices_same_radius[index_min]

        return index

    def get_cutoff_freq(self, x, y, pol):
        """
        A getter for the cutoff frequency in polarization 'pol' interpolated to arbitrary position (x, y)

        Parameters
        ----------
        x : x position (float, single value)
        y : idem for y
        pol : polarization number (int, single value)
        """
        return self.interpolators_cutoff_freq[pol](x, y)


    def __init__(self, x, y, signals, lowfreq=30.0, highfreq=500.0, sampling_period=0.1e-9, phase_method="phasor", radial_method='cubic', upsample_factor=5, coherency_cutoff_threshold=0.9, ignore_cutoff_freq_in_timing=False, verbose=False):
        """
        Initialize a callable signal interpolator object

        Parameters
        ----------
        x : 1D array for the simulated antenna positions (x) in m
        y : idem for y
        signals : 3D array of shape (Nant, Nsamples, Npols) with the antennas indexed in the first axis, the time traces in the second axis, and the polarizations in the third.
        lowfreq: low-frequency limit, typically set to 30 MHz. If a higher low-frequency limit is desired, it may likely be better to keep it at 30 MHz here, and high-pass filter later.
        highfreq : high-frequency limit, default 500.0 MHz, adjustable to e.g. 80 MHz.
        sampling_period : the time between samples in the data, default 0.1e-9 seconds (0.1 ns)
        phase_method : the options for the phase interpolation are "phasor" and "timing", cf. pg 8 in the article ("Method (1)" vs "Method (2)"), default "phasor" (Method (1))
        radial_method : the interp1d method used for radially interpolating Fourier coefficients. Usually set to 'cubic' for cubic splines, in interpolation_fourier.
        upsample_factor : upsampling factor used for sub-sample timing accuracy. Default 5.
        coherency_cutoff_threshold : the value of Eq. (2.4) that defines the reliable high-frequency limit on each position. Used internally in the "timing" method, also available to the user via self.get_cutoff_freq(x, y, pol) above. Default 0.9.
        ignore_cutoff_freq_in_timing : can be set to True when experimenting with the "timing" method without its stopping criterion, default False.
        verbose : print info while initializing
        """
        self.nofcalls = 0
        self.verbose = verbose
        self.method = phase_method
        self.pos_x = x
        self.pos_y = y
        (Nants, Nsamples, Npols) = signals.shape # hard assumption, 3D...
        self.trace_length = Nsamples
        self.sampling_period = sampling_period
        if self.verbose: print('Setting sampling period to %1.1e seconds' % self.sampling_period)
        # Get the abs-amplitude and phase spectra from the time traces
        (self.freqs, all_antennas_spectrum, self.abs_spectrum, self.phasespectrum, self.unwrapped_phases) = self.get_spectra(signals)

        if self.verbose:
            print('Doing timings using hilbert envelope...')

        # Get pulse timings using Hilbert envelope
        self.pulse_timings = self.hilbert_envelope_timing(signals, lowfreq=30.0, highfreq=80.0, upsample_factor=upsample_factor)

        # Remove timings from phase spectra
        self.phasespectrum_corrected = self.timing_corrected_phases(self.freqs, self.phasespectrum, self.pulse_timings)

        # Get phase constant or `Hilbert phase`
        self.const_phases = self.get_constant_phases(high_freq=80.0)

        # Remove phase constant from phase spectra
        self.phasespectrum_corrected -= self.const_phases[:, np.newaxis, :] # from (Nant, Npol) to (Nant, Nsamples, Npol)
        self.phasespectrum_corrected_before_freq_dependent = np.copy(self.phasespectrum_corrected) # for testing / demo purposes

        # Get degree of coherency and reliable high-cutoff frequency
        if self.verbose:
            print('Getting coherency and freq cutoff')
        (self.coherency_vs_freq, self.cutoff_freq) = self.get_coherency_vs_frequency(coherency_cutoff=coherency_cutoff_threshold)
        if self.method == "timing":
            if verbose: print('Doing freq dependent timing...')
            # Get arrival times in a sliding frequency window
            self.freq_dependent_timing = self.get_freq_dependent_timing_correction(upsample_factor=upsample_factor, ignore_cutoff_freq_in_timing=ignore_cutoff_freq_in_timing)
            if verbose: print('Done freq dependent timing')
        else:
            # print('NOT doing freq dependent timing')
            self.freq_dependent_timing = np.zeros(self.phasespectrum_corrected.shape)

        # Remove sliding-window timings from phase spectra
        for i, ant in enumerate(range(Nants)):
            for pol in range(Npols):
                this_phase_corrections = 2*np.pi * (self.freqs*1.0e6) * self.freq_dependent_timing[ant, :, pol]
                self.phasespectrum_corrected[ant, :, pol] = self.phasespectrum_corrected[ant, :, pol] + this_phase_corrections # this can be done more efficiently...

        self.coherency = self.degree_of_coherency()

        """
        Produce interpolators for amplitude and phase spectrum
        Do the Fourier interpolation for each frequency bin < 500 MHz, and for each polarization, separately
        Note: an order of magnitude speed improvement should be obtainable by vectorizing the Fourier interpolator
        """
        nof_freq_channels = len(np.where( (self.freqs < 500) )[0]) # remove hardcoded number

        self.interpolators_abs_spectrum = np.empty( (Npols, nof_freq_channels), dtype=object)# [ [None]*nof_freq_channels ] * Npols

        """
        Create interpolators for the "phasors" for each frequency,
        i.e. exp(i phi(f)) = (cos(i phi), sin(i phi)) for each frequency
        """
        self.interpolators_cosphi = np.empty( (Npols, nof_freq_channels), dtype=object)
        self.interpolators_sinphi = np.empty( (Npols, nof_freq_channels), dtype=object)

        self.interpolators_freq_dependent_timing = np.empty( (Npols, nof_freq_channels), dtype=object)

        self.interpolators_timing = np.empty(Npols, dtype=object)

        self.interpolators_constphase = np.empty(Npols, dtype=object)

        self.interpolators_cutoff_freq = np.empty(Npols, dtype=object)

        if verbose: print('Creating %d interpolators total' % (3*Npols*nof_freq_channels + 3*Npols), end=' ')

        # Create and initialize the interpolators for all quantities
        for freq_channel in range(nof_freq_channels):
            for pol in range(Npols):
                self.interpolators_abs_spectrum[pol, freq_channel] = interpF.interp2d_fourier(x, y, self.abs_spectrum[:, freq_channel, pol])
                self.interpolators_freq_dependent_timing[pol, freq_channel] = interpF.interp2d_fourier(x, y, self.freq_dependent_timing[:, freq_channel, pol])

                self.interpolators_cosphi[pol, freq_channel] = interpF.interp2d_fourier(x, y, np.cos(self.phasespectrum_corrected[:, freq_channel, pol]) )
                self.interpolators_sinphi[pol, freq_channel] = interpF.interp2d_fourier(x, y, np.sin(self.phasespectrum_corrected[:, freq_channel, pol]) )


        for pol in range(Npols):
            self.interpolators_timing[pol] = interpF.interp2d_fourier(x, y, self.pulse_timings[:, pol])
            self.interpolators_constphase[pol] = interpF.interp2d_fourier(x, y, self.const_phases[:, pol])
            self.interpolators_cutoff_freq[pol] = interpF.interp2d_fourier(x, y, self.cutoff_freq[:, pol])

        if self.verbose:
            print('Done.')
    # end __init__

    def __call__(self, x, y, lowfreq=30.0, highfreq=500.0, filter_up_to_cutoff=False, account_for_timing=True, const_time_offset=20.0e-9, full_output=False):
        """
        Call the object, which computes the interpolation at arbitrary position (x, y)

        Parameters
        ----------
        x : the x position in m (float, single value)
        y : idem for y
        lowfreq : low-frequency limit for bandpass filtering of interpolated pulse, default 30.0 MHz
        highfreq : high-frequency limit, idem, default 500.0 MHz
        filter_up_to_cutoff : set to True for low-pass filtering up to local estimated cutoff frequency, default False
        account_for_timing : set to False to have each pulse at a fixed time given by 'const_time_offset' instead of its natural arrival time. Default True
        const_time_offset : constant time offset if not using interpolated arrival times. Default 20e-9 (seconds).
        full_output : set to True to output both time series and spectra. Default False, returns only time series.
        """
        if (self.nofcalls == 0) and self.verbose:
            print('Method: %s' % self.method)
        self.nofcalls += 1

        Nfreqs = len(self.interpolators_abs_spectrum[0])
        Npols = len(self.interpolators_abs_spectrum)

        freqs = np.fft.rfftfreq(self.trace_length, d=self.sampling_period)
        freqs /= 1.0e6 # in MHz
        ## Make self.freqs (todo)

        # Set up reconstructed spectra, timings, phases at freq=0
        abs_spectrum = np.zeros( (self.trace_length//2+1, Npols) )
        phasespectrum = np.zeros( (self.trace_length//2+1, Npols) )
        timings = np.zeros(Npols)
        const_phases = np.zeros(Npols)

        if self.method == 'timing':
            index_nearest = self.nearest_antenna_index(x, y)
            print('pos x = %3.2f, y = %3.2f: nearest antenna at x = %3.2f, y = %3.2f m' % (x, y, self.pos_x[index_nearest], self.pos_y[index_nearest]))

            phasespectrum = np.copy(self.phasespectrum_corrected[index_nearest]) # COPY !!!
            """
            Do nearest-neighbor interpolation on remaining ('corrected') phases, which should be near zero, but do full interpolation on amplitude spectrum
            """
            for freq_channel in range(Nfreqs):
                for pol in range(Npols): # todo: reduce
                    thisPower = self.interpolators_abs_spectrum[pol, freq_channel](x, y)
                    #thisPhase = self.interpolators_phase[pol, freq_channel](x, y)
                    abs_spectrum[freq_channel, pol] = thisPower

            # Account for freq dependent timings, which are interpolated first to (x, y)
            for freq_channel in range(Nfreqs):
                for pol in range(Npols):

                    this_timing = self.interpolators_freq_dependent_timing[pol, freq_channel](x, y)
                    this_phaseshift = -1.0e6*self.freqs[freq_channel] * 2*np.pi * this_timing
                    phasespectrum[freq_channel, pol] += this_phaseshift


        elif self.method == 'phasor':
            for freq_channel in range(Nfreqs):
                for pol in range(Npols):
                    # Interpolate abs-amplitude spectrum and phasors
                    thisPower = self.interpolators_abs_spectrum[pol, freq_channel](x, y)
                    this_realpart = self.interpolators_cosphi[pol, freq_channel](x, y)
                    this_imagpart = self.interpolators_sinphi[pol, freq_channel](x, y)

                    thisPhase = np.angle(this_realpart + 1.0j*this_imagpart)
                    # making unit vector by dividing by abs(re**2 + im**2) and multiplying that may be significantly faster

                    abs_spectrum[freq_channel, pol] = thisPower
                    phasespectrum[freq_channel, pol] = thisPhase

        else:
            raise ValueError('Unknown reconstruction method: %s' % self.method)

        # Apply the 30-80 MHz arrival times and phase constants, each interpolated to (x, y) first
        for pol in range(Npols):
            timings[pol] = self.interpolators_timing[pol](x, y)
            const_phases[pol] = self.interpolators_constphase[pol](x, y)

            # Account for timing
            if account_for_timing:
                phase_shifts = -1.0e6*freqs * 2*np.pi * timings[pol]
                phasespectrum[:, pol] += phase_shifts
            else:
                phase_shifts = -1.0e6*freqs * 2*np.pi * const_time_offset
                phasespectrum[:, pol] += phase_shifts

            # Account for constant phase
            phasespectrum[:, pol] += const_phases[pol]

        # Wrap into (-pi, pi) where needed, to tidy up
        phasespectrum = self.phase_wrap(phasespectrum)
        """
        Set frequency channels with negative abs-amplitude to 0. This can arise sometimes when Fourier-interpolating abs-amplitudes around a circle. Throw warning when this is needed.
        """
        indices_negative = np.where(abs_spectrum < 0)
        nof_negative = len(indices_negative[0])
        nof_negative_pol0 = len(np.where(indices_negative[1]==0)[0])

        if (nof_negative > 0):
            print('warning: negative values in abs_spectrum found: %d times. Setting to zero.' % nof_negative)
            abs_spectrum[indices_negative] = 0.0
        """
        Filter to bandwidth up to local cutoff frequency if desired, otherwise up to high frequency limit
        """
        for pol in range(Npols):
            high_cutoff = self.interpolators_cutoff_freq[pol](x, y) if filter_up_to_cutoff else highfreq

            filter_indices = np.where( (freqs<lowfreq) | (freqs>high_cutoff))
            abs_spectrum[filter_indices, pol] *= 0.0

        # Produce reconstructed spectrum and time series
        spectrum = abs_spectrum * np.exp(1.0j * phasespectrum)

        timeseries = np.fft.irfft(spectrum, axis=0)

        if full_output:
            return (timeseries, abs_spectrum, phasespectrum)
        else:
            return timeseries



#!/usr/bin/env python3

from setuptools import setup

setup(name='cr_pulse_interpolator', 
      version='0.2',
      description='Full electric-field waveform interpolation for air shower simulations', 
      url='https://github.com/nu-radio/cr-pulse-interpolator',
      author='Arthur Corstanje',
      author_email='a.corstanje@astro.ru.nl', 
      license='GNU General Public License v3.0',
      install_requires=['numpy', 'scipy', 'matplotlib', 'h5py'],
      packages= ['cr_pulse_interpolator'], 
      zip_safe=False)



# Demonstration script for interpolation_Fourier.py
# Author: A. Corstanje, (a.corstanje@astro.ru.nl), 2023

import numpy as np
import matplotlib.pyplot as plt
plt.ion()
from matplotlib import cm
import interpolation_fourier as interpF

def do_plot_radial(interp_fourier, max_mode=2):
    radial_interpolator = interp_fourier.get_angular_FFT_interpolator()
    fourier = interp_fourier.get_angular_FFT()
    radial_axis = interp_fourier.get_radial_axis()

    fine_radius = np.arange(0.0, 500.0, 0.5)

    fourier_interpolated = radial_interpolator(fine_radius)

    (cosines, sines) = interpF.interp2d_fourier.cos_sin_components(fourier)
    (cosines_fine, sines_fine) = interpF.interp2d_fourier.cos_sin_components(fourier_interpolated)

    y_0 = cosines_fine[:, 0]
    y_1 = cosines_fine[:, 1]

    plt.figure() # Plot cosine modes
    for k in range(max_mode+1):
        plt.plot(radial_axis, cosines[:, k], 'o', label='cos({0} phi) mode'.format(k) if k > 0 else 'Zero mode')

    for k in range(max_mode+1):
        plt.plot(fine_radius, cosines_fine[:, k])
    plt.legend(loc='best')
    plt.xlabel('Radial distance [ m ]')
    plt.ylabel('Value')

    plt.figure() # Sine modes
    for k in range(1, max_mode+1):
        plt.plot(radial_axis, sines[:, k], 'o', label='sin({0} phi) mode'.format(k))

    for k in range(1, max_mode+1):
        plt.plot(fine_radius, sines_fine[:, k])
    plt.legend(loc='best')
    plt.xlabel('Radial distance [ m ]')
    plt.ylabel('Value')


def do_plot_angular(interp_fourier, fixed_radius, values_for_radius):
    # Plots angular interpolation at a fixed radius.
    # Inputs: instance of interp2d_fourier, the fixed radius, and 1D-array of values for that radius
    phi_steps = len(values_for_radius)
    phi_step_degrees = 360.0 / phi_steps
    raw_phi_degrees = np.linspace(0.0, 360.0 - phi_step_degrees , phi_steps)

    fine_phi = np.linspace(0.0, 2*np.pi, 1000)
    fine_points_x = fixed_radius * np.cos(fine_phi)
    fine_points_y = fixed_radius * np.sin(fine_phi)

    interp_values = interp_fourier(fine_points_x, fine_points_y)
    interp_values_truncated = interp_fourier(fine_points_x, fine_points_y, max_fourier_mode=2)

    plt.figure()
    plt.plot(raw_phi_degrees, values_for_radius, 'o', label='Values at r={0:.1f} m'.format(fixed_radius))

    plt.plot(fine_phi*180/np.pi, interp_values, label='Fourier series')
    plt.plot(fine_phi*180/np.pi, interp_values_truncated, '--', label='Up to 2nd Fourier mode')
    plt.xlabel('Phi [ deg ]')
    plt.ylabel('Value')
    plt.legend(loc='best')


fname = 'sample_data.txt'
data = np.loadtxt(fname)
(x, y, values) = data.T

### Get instance of interpolator, using given values for (x, y)
fourier_interpolator = interpF.interp2d_fourier(x, y, values)
###

# Plot radial dependence of the lowest Fourier components
do_plot_radial(fourier_interpolator)

# Plot angular interpolation at two fixed radii
all_radius = np.sqrt(x**2 + y**2)
ordering_indices = fourier_interpolator.get_ordering_indices(x, y)
radius_values = all_radius[ordering_indices][:, 0] # unique radius values
for radius_stepnr in [4, 7]:
    fixed_radius = radius_values[radius_stepnr]
    values_for_radius = values[ordering_indices][radius_stepnr, :]
    do_plot_angular(fourier_interpolator, fixed_radius, values_for_radius)

# Make color plot of f(x, y), using a meshgrid
dist_scale = 250.0
ti = np.linspace(-dist_scale, dist_scale, 1000)
XI, YI = np.meshgrid(ti, ti)

### Get interpolated values at each grid point, calling the instance of interp2d_fourier
ZI = fourier_interpolator(XI, YI)
###

# And plot it
maxp = np.max(ZI)
plt.figure()
plt.gca().pcolor(XI, YI, ZI, vmax=maxp, vmin=0, cmap=cm.jet)
plt.scatter(x, y, marker='+', s=3, color='w')
mm = cm.ScalarMappable(cmap=cm.jet)
mm.set_array([0.0, maxp])
cbar = plt.colorbar(mm)
cbar.set_label('Values of f(x, y)')
plt.xlabel('x [ m ]')
plt.ylabel('y [ m ]')
plt.xlim(-250, 250)
plt.ylim(-250, 250)
plt.gca().set_aspect('equal')
